@@ -7,6 +7,9 @@ import pandas as pd
 import re
 import textdistance
 import numpy as np 
+from crim_intervals import ImportedPiece
+import pandas as pd
+import plotly.express as px
 
 # from ipywidgets import interact, fixed
 from pyvis.network import Network
@@ -16,14 +19,12 @@ import matplotlib as mplt # OY addition 6/12/24
 class NgramColorManager:
     def __init__(self):
         self.color_map = {}  # Dictionary to store pattern -> color mappings
-        
         # Use Matplotlib's qualitative colormaps
         self.qualitative_cmaps = ['tab10', 'tab20', 'Set1', 'Set2', 'Set3', 'Accent', 'Dark2', 'Paired']
     
     def generate_distinct_colors(self, n):
         """Generate n distinct colors using Matplotlib's qualitative colormaps"""
         colors = []
-        
         # First try to use the most distinct colormap (tab10)
         cmap = mplt.cm.get_cmap('tab10')
         colors.extend([mplt.colors.to_hex(cmap(i)) for i in range(min(10, n))])
@@ -34,7 +35,6 @@ class NgramColorManager:
             for cmap_name in self.qualitative_cmaps[1:]:
                 if len(colors) >= n:
                     break
-                    
                 cmap = mplt.cm.get_cmap(cmap_name)
                 # Get the number of colors in this colormap
                 if cmap_name == 'tab20':
@@ -53,7 +53,6 @@ class NgramColorManager:
                     color = mplt.colors.to_hex(cmap(i))
                     if color not in colors:  # Avoid duplicates
                         colors.append(color)
-                
                 remaining = n - len(colors)
         
         # If we still need more colors, use HSV space with golden ratio
@@ -61,7 +60,6 @@ class NgramColorManager:
             remaining = n - len(colors)
             golden_ratio_conjugate = 0.618033988749895
             h = 0.1  # Starting hue
-            
             for _ in range(remaining):
                 h = (h + golden_ratio_conjugate) % 1.0
                 s = 0.85  # High saturation for vibrant colors
@@ -70,69 +68,90 @@ class NgramColorManager:
                 colors.append(mplt.colors.to_hex(rgb))
         
         return colors[:n]
+
+class ColorManager:
+    def __init__(self):
+        self.pattern_colors = {}
+        # Use the advanced color generator from NgramColorManager
+        self.ngram_color_manager = NgramColorManager()
+        self.color_index = 0
     
-    def get_color_for_pattern(self, pattern):
-        """Get a color for a pattern, creating a new one if it doesn't exist"""
-        pattern_str = pattern
-        if not isinstance(pattern, str):
-            pattern_str = ", ".join(str(item) for item in pattern)
-            
-        if pattern_str not in self.color_map:
-            # We'll assign colors in order as new patterns are encountered
-            n = len(self.color_map) + 1
-            colors = self.generate_distinct_colors(n)
-            self.color_map[pattern_str] = colors[-1]
-            
-        return self.color_map[pattern_str]
+    def reset(self):
+        """Reset the color manager state"""
+        self.pattern_colors = {}
+        self.color_index = 0
     
-    def assign_colors_to_dataframe(self, df, pattern_column='pattern'):
-        """Assign colors to a dataframe based on patterns"""
-        # Check if dataframe is not empty
-        if len(df) > 0:
-            # Convert patterns to strings if they're not already
-            if not isinstance(df[pattern_column].iloc[0], str):
-                df = df.copy()
-                df[pattern_column] = df[pattern_column].map(
-                    lambda cell: ", ".join(str(item) for item in cell), 
-                    na_action='ignore'
-                )
-            
-            # Assign colors
-            df['color'] = df[pattern_column].apply(self.get_color_for_pattern)
+    def pre_register_patterns(self, patterns):
+        """
+        Pre-register patterns with the color manager to ensure consistency.
         
+        Args:
+            patterns: List of patterns (tuples, lists, or strings) to register
+        """
+        for pattern in patterns:
+            if isinstance(pattern, (list, tuple)):
+                pattern_str = ", ".join(str(item) for item in pattern)
+            else:
+                pattern_str = str(pattern)
+            self.get_color_for_pattern(pattern_str)
+    
+    def get_color_for_pattern(self, pattern_str):
+        """Get or assign a color for a pattern"""
+        if pattern_str not in self.pattern_colors:
+            # Get all unique patterns we need colors for
+            all_patterns = list(self.pattern_colors.keys()) + [pattern_str]
+            n_colors_needed = len(all_patterns)
+            
+            # Generate enough colors using the advanced color generator
+            colors = self.ngram_color_manager.generate_distinct_colors(n_colors_needed)
+            
+            # Assign the new color to the new pattern
+            self.pattern_colors[pattern_str] = colors[-1]
+        
+        return self.pattern_colors[pattern_str]
+    
+    def assign_colors_to_dataframe(self, df):
+        """Assign colors to dataframe patterns"""
+        df = df.copy()
+        df['color'] = df['pattern'].apply(self.get_color_for_pattern)
         return df
 
-# Create a global instance of the color manager
-color_manager = NgramColorManager()
+# Initialize global color manager
+color_manager = ColorManager()
 
 def create_bar_chart(variable, count, color, data, condition, *selectors):
+    """
+    Create a bar chart using Altair.
+    """
     color_scale = alt.Scale(
-        domain = data['pattern'].unique(),
-        range = data['color'].unique()
-    ) # OY addition 6/12/24
-
+        domain=data['pattern'].unique(),
+        range=data['color'].unique()
+    )
+    
     observer_chart = alt.Chart(data).mark_bar().encode(
         x=variable,
         y=count,
-        color=alt.Color('pattern:N', scale=color_scale, legend=alt.Legend(title='Pattern')), # OY change 6/12/24
+        color=alt.Color('pattern:N', scale=color_scale, legend=alt.Legend(title='Pattern')),
         opacity=alt.condition(condition, alt.value(1), alt.value(0.2))
     ).add_params(
         *selectors
     )
     return observer_chart
 
-
 def create_heatmap(x, x2, y, color, data, heat_map_width, heat_map_height, selector_condition, *selectors, tooltip):
+    """
+    Create a heatmap using Altair.
+    """
     color_scale = alt.Scale(
-        domain = data['pattern'].unique(),
-        range = data['color'].unique()
-    ) # OY addition 6/12/24
-
+        domain=data['pattern'].unique(),
+        range=data['color'].unique()
+    )
+    
     heatmap = alt.Chart(data).mark_bar().encode(
         x=x,
         x2=x2,
         y=y,
-        color=alt.Color('pattern:N', scale=color_scale, legend=alt.Legend(title='Pattern')), # OY change 6/12/24
+        color=alt.Color('pattern:N', scale=color_scale, legend=alt.Legend(title='Pattern')),
         opacity=alt.condition(selector_condition, alt.value(1), alt.value(0.2)),
         tooltip=tooltip
     ).properties(
@@ -143,46 +162,27 @@ def create_heatmap(x, x2, y, color, data, heat_map_width, heat_map_height, selec
     )
     return heatmap
 
-
 def _process_ngrams_df_helper(ngrams_df, main_col):
     """
     The output from the getNgram is usually a table with
     four voices and ngram of notes properties (duration or
     pitch). This method stack this property onto one column
     and mark which voices they are from.
-    :param ngrams_df: direct output from getNgram with 1 columns
-    for each voices and ngrams of notes' properties.
-    :param main_col: the name of the property
-    :return: a dataframe with ['start', main_col, 'voice'] as columns
     """
     # copy to avoid changing original ngrams df
     ngrams_df = ngrams_df.copy()
-
     # add a start column containing offsets
     ngrams_df.index.name = "start"
     ngrams_df = ngrams_df.reset_index().melt(id_vars=["start"], value_name=main_col, var_name="voice")
-
     ngrams_df["start"] = ngrams_df["start"].astype(float)
     return ngrams_df
 
-
 def process_ngrams_df(ngrams_df, ngrams_duration=None, selected_pattern=None, voices=None):
     """
     This method combines ngrams from all voices in different columns
     into one column and calculates the starts and end points of the
-    patterns. It could also filter out specific voices or patterns
-    for the users to analyze.
-
-    :param ngrams_df: dataframe we got from getNgram in crim-interval
-    :param ngrams_duration: if not None, simply output the offsets of the
-    ngrams. If we have durations, calculate the end by adding the offsets and
-    the durations.
-    :param selected_pattern: list of specific patterns the users want (optional)
-    :param voices: list of specific voices the users want (optional)
-    :return a new, processed dataframe with only desired patterns from desired voices
-    combined into one column with start and end points
+    patterns.
     """
-
     ngrams_df = _process_ngrams_df_helper(ngrams_df, 'pattern')
     if ngrams_duration is not None:
         ngrams_duration = _process_ngrams_df_helper(ngrams_duration, 'duration')
@@ -190,21 +190,18 @@ def process_ngrams_df(ngrams_df, ngrams_duration=None, selected_pattern=None, vo
     else:
         # make end=start+1 just to display offsets
         ngrams_df['end'] = ngrams_df['start'] + 1
-
+    
     # filter according to voices and patterns (after computing durations for correct offsets)
     if voices:
         voice_condition = ngrams_df['voice'].isin(voices)
         ngrams_df = ngrams_df[voice_condition].dropna(how='all')
-
     if selected_pattern:
         pattern_condition = ngrams_df['pattern'].isin(selected_pattern)
         ngrams_df = ngrams_df[pattern_condition].dropna(how='all')
-
     return ngrams_df
 
-# OY addition - new function for generating distinct colors 6/12/24
 def generate_distinct_colors(n):
-    # Generate `n` distinct colors using the HSV color space
+    """Generate `n` distinct colors using the HSV color space"""
     colors = []
     for i in range(n):
         hue = i / n
@@ -214,160 +211,120 @@ def generate_distinct_colors(n):
         colors.append(color)
     return colors
 
-# OY addition - new function for adding colors 6/12/24
-# def ngrams_color_helper(new_processed_ngrams_df: pd.DataFrame) -> pd.DataFrame:
-#     """
-#     Add a Series to the dataframe that assigns a unique hex color to each unique pattern
-#     :param new_processed_ngrams_df: processed crim-intervals getNgram's output where tuples have been converted to strings
-#     :return: a dataframe containing a new column with hex color values
-#     """
-
-#     # Calculate the number of unique values in 'pattern_str'
-#     num_unique_values = new_processed_ngrams_df['pattern'].nunique() #  Function to count unique values in 'pattern_str' column
-#     # # Generate enough hex colors
-#     hex_colors = generate_distinct_colors(num_unique_values)   
-#     # # Step 2: Assign colors to unique values
-#     color_map = dict(zip(new_processed_ngrams_df['pattern'].unique(), hex_colors)) 
-#     # # Add a new column 'color' to the DataFrame
-#     new_processed_ngrams_df['color'] = new_processed_ngrams_df['pattern'].map(color_map)
-
-#     return new_processed_ngrams_df
-
-# def _plot_ngrams_df_heatmap(processed_ngrams_df, heatmap_width=800, heatmap_height=300, includeCount=False):
-#     """
-#     Plot a heatmap for crim-intervals getNgram's processed output.
-#     :param ngrams_df: processed crim-intervals getNgram's output.
-#     :param selected_pattern: list of specific patterns the users want (optional)
-#     :param voices: list of specific voices the users want (optional)
-#     :param heatmap_width: the width of the final heatmap (optional)
-#     :param heatmap_height: the height of the final heatmap (optional)
-#     :return: a bar chart that displays the different patterns and their counts,
-#     and a heatmap with the start offsets of chosen voices / patterns
-#     """
-
-#     processed_ngrams_df = processed_ngrams_df.dropna(how='any')
-#     selector = alt.selection_point(fields=['pattern'])
-#     y = alt.Y("voice", sort=None)
-
-#     # make a copy of the processed n_grams and turn them into Strings
-#     new_processed_ngrams_df = processed_ngrams_df.copy()
-#     new_processed_ngrams_df['pattern'] = processed_ngrams_df['pattern'].map(lambda cell: ", ".join(str(item) for item in cell), na_action='ignore')
-
-#     new_processed_ngrams_df = ngrams_color_helper(new_processed_ngrams_df) # OY addition 6/12/24
-    
-#     heatmap = create_heatmap('start', 'end', y, 'pattern', new_processed_ngrams_df, heatmap_width, heatmap_height,
-#                              selector, selector, tooltip=['start', 'end', 'pattern'])
-#     if includeCount:
-#         variable = alt.X('pattern', axis=alt.Axis(labelAngle=-45))
-#         patterns_bar = create_bar_chart(variable, 'count(pattern)', 'pattern', new_processed_ngrams_df, selector, selector)
-#         return alt.vconcat(patterns_bar, heatmap)
-#     else:
-#         return heatmap
-
-# Modified version of your ngrams_color_helper function
 def ngrams_color_helper(new_processed_ngrams_df: pd.DataFrame) -> pd.DataFrame:
     """
     Add a Series to the dataframe that assigns a unique hex color to each unique pattern
     using the global color manager to ensure consistency across plots.
-    
-    :param new_processed_ngrams_df: processed crim-intervals getNgram's output where tuples have been converted to strings
-    :return: a dataframe containing a new column with hex color values
     """
     return color_manager.assign_colors_to_dataframe(new_processed_ngrams_df)
 
-def _plot_ngrams_df_heatmap(processed_ngrams_df, heatmap_width=800, heatmap_height=300, includeCount=False):
+def _plot_ngrams_df_heatmap(processed_ngrams_df, heatmap_width=800, heatmap_height=300, includeCount=False, title=None):
     """
     Plot a heatmap for crim-intervals getNgram's processed output.
-    :param ngrams_df: processed crim-intervals getNgram's output.
-    :param selected_pattern: list of specific patterns the users want (optional)
-    :param voices: list of specific voices the users want (optional)
-    :param heatmap_width: the width of the final heatmap (optional)
-    :param heatmap_height: the height of the final heatmap (optional)
-    :return: a bar chart that displays the different patterns and their counts,
-    and a heatmap with the start offsets of chosen voices / patterns
     """
     processed_ngrams_df = processed_ngrams_df.dropna(how='any')
     selector = alt.selection_point(fields=['pattern'])
     y = alt.Y("voice", sort=None)
+    
+    # make a copy of the processed n_grams and turn them into Strings
+    new_processed_ngrams_df = processed_ngrams_df.copy()
+    new_processed_ngrams_df['pattern'] = processed_ngrams_df['pattern'].map(
+        lambda cell: ", ".join(str(item) for item in cell),
+        na_action='ignore'
+    )
+    
+    # Use the color manager to assign consistent colors
+    new_processed_ngrams_df = ngrams_color_helper(new_processed_ngrams_df)
+    
+    heatmap = create_heatmap('start', 'end', y, 'pattern', new_processed_ngrams_df, heatmap_width, heatmap_height,
+                            selector, selector, tooltip=['start', 'end', 'pattern'])
+    
+    if includeCount:
+        variable = alt.X('pattern', axis=alt.Axis(labelAngle=-45))
+        patterns_bar = create_bar_chart(variable, 'count(pattern)', 'pattern', new_processed_ngrams_df, selector, selector)
+        chart = alt.vconcat(patterns_bar, heatmap)
+    else:
+        chart = heatmap
+    
+    # Apply title if provided
+    if title is not None:
+        chart = chart.properties(title=title)
+    
+    return chart
 
+def plot_ngrams_heatmap(ngrams_df, ngrams_duration=None, selected_patterns=[], voices=[], 
+                       heatmap_width=800, heatmap_height=300, includeCount=False, title=None):
+    """
+    Plot a heatmap for crim-intervals getNgram's output.
+    """
+    processed_ngrams_df = process_ngrams_df(ngrams_df, ngrams_duration=ngrams_duration,
+                                          selected_pattern=selected_patterns, voices=voices)
+    return _plot_ngrams_df_heatmap(processed_ngrams_df, heatmap_width=heatmap_width,
+                                 heatmap_height=heatmap_height, includeCount=includeCount, title=title)
+
+def plot_ngrams_barchart(ngrams_df, ngrams_duration=None, selected_patterns=[], voices=[], chart_width=800,
+                        chart_height=300):
+    """
+    Plot a bar chart for crim-intervals getNgram's output.
+    """
+    processed_ngrams_df = process_ngrams_df(ngrams_df, ngrams_duration=ngrams_duration,
+                                          selected_pattern=selected_patterns,
+                                          voices=voices)
+    return _plot_ngrams_df_barchart(processed_ngrams_df, chart_width=chart_width, chart_height=chart_height)
+
+def _plot_ngrams_df_barchart(processed_ngrams_df, chart_width=800, chart_height=300):
+    """
+    Plot a bar chart for crim-intervals getNgram's processed output.
+    """
+    processed_ngrams_df = processed_ngrams_df.dropna(how='any')
+    selector = alt.selection_point(fields=['pattern'])
+    
+    # make a copy of the processed n_grams and turn them into Strings
+    new_processed_ngrams_df = processed_ngrams_df.copy()
+    new_processed_ngrams_df['pattern'] = processed_ngrams_df['pattern'].map(
+        lambda cell: ", ".join(str(item) for item in cell),
+        na_action='ignore'
+    )
+    
+    # Add color assignment for consistency
+    new_processed_ngrams_df = ngrams_color_helper(new_processed_ngrams_df)
+    
+    variable = alt.X('pattern', axis=alt.Axis(labelAngle=-45))
+    return create_bar_chart(variable, 'count(pattern)', 'pattern', new_processed_ngrams_df, selector, selector)
+
+
+# Helper function for your analysis code
+def pre_register_patterns(shared_entries):
+    """Pre-register patterns with the color manager to ensure consistency"""
+    for pattern in shared_entries:
+        if isinstance(pattern, (list, tuple)):
+            pattern_str = ", ".join(str(item) for item in pattern)
+        else:
+            pattern_str = str(pattern)
+        color_manager.get_color_for_pattern(pattern_str)
+# 2025 version
+# FIXED: Updated bar chart function to include color consistency
+def _plot_ngrams_df_barchart(processed_ngrams_df, chart_width=800, chart_height=300):
+    """
+    Plot a bar chart for crim-intervals getNgram's processed output.
+    """
+    processed_ngrams_df = processed_ngrams_df.dropna(how='any')
+    selector = alt.selection_point(fields=['pattern'])
+    
     # make a copy of the processed n_grams and turn them into Strings
     new_processed_ngrams_df = processed_ngrams_df.copy()
     new_processed_ngrams_df['pattern'] = processed_ngrams_df['pattern'].map(
         lambda cell: ", ".join(str(item) for item in cell), 
         na_action='ignore'
     )
-
-    # Use the color manager to assign consistent colors
+    
+    # FIXED: Add color assignment for consistency
     new_processed_ngrams_df = ngrams_color_helper(new_processed_ngrams_df)
     
-    heatmap = create_heatmap('start', 'end', y, 'pattern', new_processed_ngrams_df, heatmap_width, heatmap_height,
-                             selector, selector, tooltip=['start', 'end', 'pattern'])
-    if includeCount:
-        variable = alt.X('pattern', axis=alt.Axis(labelAngle=-45))
-        patterns_bar = create_bar_chart(variable, 'count(pattern)', 'pattern', new_processed_ngrams_df, selector, selector)
-        return alt.vconcat(patterns_bar, heatmap)
-    else:
-        return heatmap
-
-def plot_ngrams_heatmap(ngrams_df, ngrams_duration=None, selected_patterns=[], voices=[], heatmap_width=800,
-                        heatmap_height=300, includeCount=False):
-    """
-    Plot a heatmap for crim-intervals getNgram's output.
-    :param ngrams_df: crim-intervals getNgram's output
-    :param ngrams_duration: if not None, rely on durations in the
-    df to calculate the durations of the ngrams.
-    :param selected_patterns: list of specific patterns the users want (optional)
-    :param voices: list of specific voices the users want (optional)
-    :param heatmap_width: the width of the final heatmap (optional)
-    :param heatmap_height: the height of the final heatmap (optional)
-    :return: a bar chart that displays the different patterns and their counts,
-    and a heatmap with the start offsets of chosen voices / patterns
-    """
-    processed_ngrams_df = process_ngrams_df(ngrams_df, ngrams_duration=ngrams_duration,
-                                            selected_pattern=selected_patterns,
-                                            voices=voices)
-    return _plot_ngrams_df_heatmap(processed_ngrams_df, heatmap_width=heatmap_width, heatmap_height=heatmap_height, includeCount=includeCount)
-
-def plot_ngrams_barchart(ngrams_df, ngrams_duration=None, selected_patterns=[], voices=[], chart_width=800,
-                        chart_height=300):
-    """
-    Plot a bar chart for crim-intervals getNgram's output.
-    :param ngrams_df: crim-intervals getNgram's output
-    :param ngrams_duration: if not None, rely on durations in the
-    df to calculate the durations of the ngrams.
-    :param selected_patterns: list of specific patterns the users want (optional)
-    :param voices: list of specific voices the users want (optional)
-    :param chart_width: the width of the final bar chart (optional)
-    :param chart_height: the height of the final bar chart (optional)
-    :return: a bar chart that displays the different patterns and their counts
-    """
-    processed_ngrams_df = process_ngrams_df(ngrams_df, ngrams_duration=ngrams_duration,
-                                            selected_pattern=selected_patterns,
-                                            voices=voices)
-    return _plot_ngrams_df_barchart(processed_ngrams_df, chart_width=chart_width, chart_height=chart_height)
-
-def _plot_ngrams_df_barchart(processed_ngrams_df, chart_width=800, chart_height=300):
-    """
-    Plot a bar chart for crim-intervals getNgram's processed output.
-    :param ngrams_df: processed crim-intervals getNgram's output.
-    :param selected_pattern: list of specific patterns the users want (optional)
-    :param voices: list of specific voices the users want (optional)
-    :param chart_width: the width of the final bar chart (optional)
-    :param chart_height: the height of the final bar chart (optional)
-    :return: a bar chart that displays the different patterns and their counts
-    """
-
-    processed_ngrams_df = processed_ngrams_df.dropna(how='any')
-    selector = alt.selection_point(fields=['pattern'])
-    y = alt.Y("voice", sort=None)
-
-    # make a copy of the processed n_grams and turn them into Strings
-    new_processed_ngrams_df = processed_ngrams_df.copy()
-    new_processed_ngrams_df['pattern'] = processed_ngrams_df['pattern'].map(lambda cell: ", ".join(str(item) for item in cell), na_action='ignore')
-
     variable = alt.X('pattern', axis=alt.Axis(labelAngle=-45))
     return create_bar_chart(variable, 'count(pattern)', 'pattern', new_processed_ngrams_df, selector, selector)
 
+# unchanged
 def _from_ema_to_offsets(df, ema_column):
     """
     This method adds a columns of start and end measure of patterns into
@@ -684,3 +641,207 @@ def _manipulate_processed_network_df(df, interval_column, search_pattern_starts_
 #     networks_dict = create_interval_networks(df[interval_column], interval_type)
 #     df = process_network_df(df, interval_column, ema_column)
 #     return networks_dict, create_interactive_compare_df(df, interval_column)
+
+# new cadence visualization functions
+
+
+# cadence radar
+def cadence_radar(cadences, groupby_column='Title', chart_title="Radar Plot of Cadence Tones"):
+    """
+    Create a radar plot showing distribution of cadence tones.
+    The optional chart_title parameter allows for a custom title.
+
+    Cadences df must contain 'Composer', 'Title' and 'Tone' columns.  Groupby colum
+
+    Parameters:
+    -----------
+    cadences : pd.DataFrame
+        DataFrame containing cadence data
+
+    groupby_column : str
+        Column name to group by (default is 'Title'); could also be 'Composer'
+    chart_title : str
+        Title for the chart (default is "Radar Plot of Cadence Tones")  
+
+
+
+    Returns:
+    --------
+    fig : go.Figure
+        Plotly Figure object representing the radar plot
+
+    Notes:
+    ------
+    This function creates a radar plot to visualize the relative distribution of cadence tones across different titles in the corpus.
+    It uses Plotly Express to generate the plot and applies various customizations such as color mapping, legend positioning, and layout adjustments.
+    The plot displays the percentage of occurrence for each tone within each title, allowing for easy comparison between titles and tones.
+    """
+    # Define constants at function scope
+    tone_order = {'C': 0, 'D': 1, 'E-': 2, 'E': 3, 'F': 4,
+        'G': 5, 'A': 6, 'B-': 7
+    }
+
+    
+    # Vectorized approach using groupby operations
+    grouped = cadences.groupby([groupby_column, 'Tone']).size().reset_index(name='count')
+    
+    # Calculate percentages efficiently
+    title_sums = grouped.groupby(groupby_column)['count'].sum()
+    grouped['Percentage'] = (grouped['count'] / grouped[groupby_column].map(title_sums)) * 100
+    
+    # Create radar plot
+    fig = px.line_polar(
+        grouped[grouped['count'] > 0],
+        r='Percentage',
+        theta='Tone',
+        line_close=True,
+        color=groupby_column,
+        markers=True,
+        category_orders={'Tone': sorted(tone_order.keys(), key=lambda x: tone_order[x])}
+    )
+    
+    # Apply optimizations
+    fig.update_traces(fill='toself', line=dict(width=2))
+    
+    # Update layout with bottom legend
+    fig.update_layout(
+        width=800,
+        height=600,
+        legend=dict(
+            orientation="h",  # Horizontal layout for better space usage
+            yanchor="bottom",
+            y=-0.4,           # Position below plot
+            xanchor="center",
+            x=0.5,            # Center horizontally
+            xref="container", # Scale with container
+            yref="container", # Scale with container
+            title=dict(
+                text=groupby_column,
+                side="top",
+                font_size=12
+            ),
+            itemsizing='constant',
+            itemwidth=30,
+            bordercolor="black",
+            borderwidth=1,
+            bgcolor="rgba(255,255,255,0.8)"
+        ),
+        title=dict(text=chart_title, x=0.5),
+        polar=dict(
+            radialaxis=dict(visible=True, title='Percentage')
+        ),
+        margin=dict(b=120)  # Bottom margin for legend
+    )
+    
+    return fig
+
+
+
+# cadence progress viz
+def cadence_progress(cadences, chart_title="Radar Plot of Cadence Tones"):
+    """
+    Create a scatter plot showing progress of cadences in a composition.
+    Cadence df must contain 'Composer', 'Title', 'Progress', 'Tone', and 'CadType' columns.
+
+    Parameters:
+    -----------
+    cadences : pd.DataFrame
+        DataFrame containing cadence data
+    chart_title : str
+        Title for the chart (default is "Radar Plot of Cadence Tones")
+
+    Returns:
+    --------
+    fig : go.Figure
+        Plotly Figure object representing the scatter plot
+
+    Notes:
+    ------
+    This function generates a scatter plot to visualize the progression of cadences in a specific composition.
+    It categorizes cadences based on their type and assigns distinct colors to each cadence type.
+    The plot displays the cadence progression along the x-axis and the tone along the y-axis, allowing for easy identification of cadence changes throughout the composition.
+    Custom color mappings are used to differentiate between original and modified cadence types.
+    The plot includes a horizontal legend for easy identification of cadence types.
+    """
+    # Define custom ordering for CadTypes
+    custom_order = ['Authentic', 'Evaded Authentic', 
+        'Clausula Vera', 'Evaded Clausula Vera', 'Abandoned Clausula Vera', 'Phrygian Clausula Vera',
+        'Double Leading Tone', 'Evaded Double Leading Tone', 'Abandoned Double Leading Tone',
+        'Altizans Only', 'Phrygian Altizans', 'Evaded Altizans Only',
+        'Phrygian', 'Reinterpreted', 'Quince', 'Leaping Contratenor']
+
+    # order for y axis tones:
+    custom_tone_order = ['E-', 'B-', 'F', 'C', 'G', 'D', 'A', 'E', 'B']  
+
+    # color map for cadences--the evaded types are less saturated than corresponding regular types
+    color_mapping = {'Authentic': '#FF0000',
+     'Clausula Vera': '#0000FF',
+     'Double Leading Tone': '#00FF00',
+     'Altizans Only': '#800080',
+     'Phrygian': '#00FF00',
+     'Reinterpreted': '#00FFFF',
+     'Quince': '#4B0082',
+     'Leaping Contratenor': '#000000',
+     'Phrygian Clausula Vera': '#FF00FF',
+     'Evaded Authentic': '#FF6666',
+     'Evaded Clausula Vera': '#6666FF',
+     'Evaded Double Leading Tone': '#66FF66',
+     'Evaded Altizans Only': '#CC99CC',
+     'Evaded Phrygian': '#66FF66',
+     'Evaded Reinterpreted': '#99FFFF',
+     'Evaded Quince': '#9966CC',
+     'Evaded Phrygian Clausula Vera': '#FF66FF',
+     'Abandoned Clausula Vera': '#CCCCFF',
+     'Abandoned Double Leading Tone': '#CCFFCC',
+     'Abandoned Phrygian': '#FFE5CC',
+     'Abandoned Authentic': '#FFCCCC',
+     'Abandoned Phrygian Clausula Vera': '#FFCCFF'}
+    
+    composer = cadences['Composer'].iloc[0]
+    title = cadences['Title'].iloc[0]
+
+    # Ensure the Tone column is categorical with correct ordering
+    cadences['Tone'] = pd.Categorical(
+        cadences['Tone'],
+        categories=custom_tone_order,  # specify tone order for y axis
+        ordered=True
+    )
+    
+    # Create figure with explicit category ordering
+    fig = px.scatter(
+        cadences,
+        x='Progress',
+        y='Tone',
+        color='CadType',
+        color_discrete_map=color_mapping, # use color mapping from above
+        category_orders={'Tone': custom_tone_order}
+    )
+    
+    # Configure Y-axis to show E- at bottom
+    fig.update_layout(
+        yaxis=dict(
+            categoryorder='array',
+            categoryarray=custom_tone_order,  # specify tone order for y axis
+            autorange=True,
+            fixedrange=True,
+            scaleanchor='y',
+            scaleratio=1,
+            showgrid=True,
+            showticklabels=True
+        ),
+        yaxis_range=[None, None]  # Allow Plotly to determine range
+    )
+    
+    # Update marker properties
+    fig.update_traces(marker=dict(size=25))
+    
+    # Customize layout
+    fig.update_layout(
+        title=f'Progress of Cadences in {composer}: {title}',
+        legend=dict(
+            title_text="Cadence Type",
+        orientation="v"
+        )
+    )
+    
+    return fig
